use json::JsonValue;
use serde_derive::Deserialize;
use serde_derive::Serialize;
use serde_json::Value;
use strum_macros::Display;
use std::{time::{Duration, Instant}, path::Display, sync::Arc};
use getset::{Getters, Setters};
use reqwest::{blocking::get, StatusCode};

#[derive(Default, Debug, Clone, PartialEq, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct DiscordInvite {
    #[serde(rename = "type")]
    pub type_field: i64,
    pub code: String,
    pub inviter: Option<Inviter>,
    #[serde(rename = "expires_at")]
    pub expires_at: Option<chrono::NaiveDateTime>,
    pub flags: Option<i64>,
    pub guild: Option<Guild>,
    #[serde(rename = "guild_id")]
    pub guild_id: String,
    pub channel: Option<Channel>,
    #[serde(rename = "approximate_member_count")]
    pub approximate_member_count: i64,
    #[serde(rename = "approximate_presence_count")]
    pub approximate_presence_count: i64,
}

#[derive(Default, Debug, Clone, PartialEq, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Inviter {
    pub id: Option<String>,
    pub username: Option<String>,
    pub avatar: Option<String>,
    pub discriminator: Option<String>,
    #[serde(rename = "public_flags")]
    pub public_flags: Option<i64>,
    #[serde(rename = "premium_type")]
    pub premium_type: Option<i64>,
    pub flags: Option<i64>,
    pub banner: Option<String>,
    #[serde(rename = "accent_color")]
    pub accent_color: Option<i64>,
    #[serde(rename = "global_name")]
    pub global_name: Option<String>,
    #[serde(rename = "avatar_decoration_data")]
    pub avatar_decoration_data: Option<Value>,
    #[serde(rename = "banner_color")]
    pub banner_color: Option<String>,
}

#[derive(Default, Debug, Clone, PartialEq, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Guild {
    pub id: Option<String>,
    pub name: Option<String>,
    pub splash: Option<String>,
    pub banner: Option<String>,
    pub description: Option<Value>,
    pub icon: Option<String>,
    pub features: Option<Vec<String>>,
    #[serde(rename = "verification_level")]
    pub verification_level: Option<i64>,
    #[serde(rename = "vanity_url_code")]
    pub vanity_url_code: Option<String>,
    #[serde(rename = "nsfw_level")]
    pub nsfw_level: Option<i64>,
    pub nsfw: Option<bool>,
    #[serde(rename = "premium_subscription_count")]
    pub premium_subscription_count: Option<i64>,
}

#[derive(Default, Debug, Clone, PartialEq, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Channel {
    pub id: Option<String>,
    #[serde(rename = "type")]
    pub type_field: Option<i64>,
    pub name: Option<String>,
}

impl DiscordInvite {
    pub fn resolve(code: String) -> Result<Self, DiscordInviteError>{
        let response = match get(format!("https://discord.com/api/v9/invites/{}?with_counts=true&with_expiration=true", code)){
            Ok(r) => r,
            Err(e) => return Err(DiscordInviteError::Unknown(format!("Unknown Error!: {}", e))),
        };
        let result = match response.text(){
            Ok(r) => r,
            Err(e) => return Err(DiscordInviteError::Unknown(format!("Unknown Error!: {}", e))),
        };
        if result.contains("Unknown Invite"){
            return Err(DiscordInviteError::InvalidInvite);
        }
        if result.contains("You are being rate limited."){
            let time: Value = match serde_json::from_str(&result){
                Ok(v) => v,
                Err(_) => return Err(DiscordInviteError::RateLimit(Duration::ZERO)),
            };
            let time = match time.get("retry_after"){
                Some(v) => v,
                None => return Err(DiscordInviteError::RateLimit(Duration::ZERO)),
            }.as_f64().unwrap_or(0.0);
            return Err(DiscordInviteError::RateLimit(Duration::from_secs_f64(time)));
        }
        if result.contains("Error 1015"){
            return Err(DiscordInviteError::RateLimit(Duration::ZERO));
        }
        let parsed_result = match serde_json::from_str::<DiscordInvite>(&result){
            Ok(r) => r,
            Err(e) => {
                return Err(DiscordInviteError::Unknown(e.to_string()));
            }
        };
        Ok(parsed_result)
    }
}

#[derive(Debug)]
pub enum DiscordInviteError {
    InvalidInvite,
    RateLimit(Duration),
    Timeout,
    Unknown(String),
}

impl std::fmt::Display for DiscordInviteError{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", match self {
            Self::Unknown(s) => format!("Unknown Error!: {}", s),
            Self::InvalidInvite => "Invalid Invite".to_string(),
            Self::RateLimit(t) => format!("Youve been rate limited for {}s", t.as_secs_f64()),
            _ => "Not yet implemented".to_string()
        })
    }
}